#lang scheme
(require "test.ss")

;; A success is (any stream -> alpha)
;; A fail is (-> alpha)

(define table
  `([main [("apple") ,(lambda () 'apple)]
          [("banana") ,(lambda () 'banana)]
          [(fruit-salad "!") ,(lambda (x)
                                `(,x !))]]
    [fruit-salad [(fruit fruit) ,(lambda (x y)
                                   (list x y))]
                 [(fruit fruit fruit) ,(lambda (x y z)
                                         (list x y z))]]
    [fruit [("orange") ,(lambda () 'orange)]
           [("grape") ,(lambda () 'grape)]]))

;; parse : sym table list-of-string success fail -> alpha
(define (parse what table stream 
               success-k fail-k)
  (local [(define defn (assq what table))]
    (if defn
        (parse-alts (rest defn) table stream 
                    success-k fail-k)
        (error 'parse "no definition for ~a" what))))

(define (ok v stream) (cons v stream))
(define (nope) #f)

(test (parse 'main table '("apple") ok nope) (cons 'apple '()))
(test (parse 'fruit-salad table '("orange" "orange") ok nope) 
      (cons '(orange orange) '()))
(test (parse 'main table '("orange" "orange" "!") ok nope)
      (cons '((orange orange) !) '()))
#;(test (parse 'main table '("orange" "orange" "orange" "!") ok nope)
        (cons '((orange orange orange) !) '()))
(test (parse 'main table '("radish") ok nope) #f)

;; parse-alts : list-of-case table list-of-string success fail
;;              -> alpha
(define (parse-alts alts table stream success-k fail-k)
  (if (null? alts)
      (fail-k)
      (parse-case (first alts) table stream
                  success-k
                  (lambda ()
                    (parse-alts (rest alts) table stream
                                success-k fail-k)))))

(let ([alts (list
             `[("apple") ,(lambda () 'apple)]
             `[("grape") ,(lambda () 'grape)])])
  (test (parse-alts alts table '("apple") ok nope) (cons 'apple '()))
  (test (parse-alts alts table '("grape") ok nope) (cons 'grape '()))
  (test (parse-alts alts table '("grape" "fish") ok nope)
        (cons 'grape '("fish")))
  (test (parse-alts alts table '("radish") ok nope) #f))

;; parse-case : case table list-of-string success fail
;;              -> alpha
(define (parse-case case table stream success-k fail-k)
  (local [(define seq (first case))
          (define handler (second case))]
    (parse-sequence seq table stream
                    (lambda (vals stream)
                      (success-k (apply handler vals) stream))
                    fail-k)))

(let ([case `[("apple") ,(lambda () 'apple)]])
  (test (parse-case case table '("apple") ok nope) (cons 'apple '()))
  (test (parse-case case table '("grape") ok nope) #f))

;; parse-sequence : sequence table list-of-string success fail
;;                  -> alpha
(define (parse-sequence seq table stream success-k fail-k)
  (cond
    [(empty? seq) (success-k empty stream)]
    [else 
     (parse-item (first seq) table stream
                 (lambda (fst-vals stream)
                   (parse-sequence (rest seq) table stream
                                   (lambda (snd-vals stream)
                                     (success-k (append fst-vals
                                                        snd-vals)
                                                stream))
                                   fail-k))
                 fail-k)]))

(let ([seq `("apple" "grape")])
  (test (parse-sequence '("apple") table '("apple") ok nope) 
        (cons empty '()))
  (test (parse-sequence seq table '("apple" "grape") ok nope)
        (cons empty '()))
  (test (parse-sequence seq table '("apple" "radish") ok nope) #f))
(let ([seq '(fruit fruit)])
  (test (parse-sequence seq table '("orange" "grape") ok nope)
        (cons '(orange grape) '()))
  (test (parse-sequence seq table '("grape" "orange") ok nope)
        (cons '(grape orange) '())))
(let ([seq '(fruit-salad "!")])
  (test (parse-sequence seq table '("orange" "orange" "!") ok nope)
        (cons '((orange orange)) '()))
  #;(test (parse-sequence seq table '("orange" "orange" "orange" "!") ok nope)
          (cons '((orange orange orange)) '())))

;; parse-item : sequence table list-of-string success fail
;;              -> alpha
(define (parse-item item table stream success-k fail-k)
  (cond
    [(string? item)
     (if (and (cons? stream)
              (equal? item (first stream)))
         (success-k empty (rest stream))
         (fail-k))]
    [else
     (parse item table stream
            (lambda (v stream)
              (success-k (list v) stream))
            fail-k)]))

(test (parse-item '"grape" table '("grape") ok nope)
      (cons empty '()))
(test (parse-item 'fruit table '("grape") ok nope)
      (cons '(grape) '()))
(test (parse-item 'fruit table '("radish") ok nope) #f)

(run-tests)
